@@ -1,67 +1,60 @@
 """Descriptors for binary records."""
 
+import copy
 import enum
 import math
 import types
 import warnings
 import dataclasses
-from typing import Optional, Iterable, Type, Union
+from typing import Iterable, Optional, Sequence, Type, Union
 
+import bpack.typing
 import bpack.utils
 from .utils import classdecorator
-
+from .enums import EBaseUnits, EByteOrder, EBitOrder
 
 __all__ = [
-    'descriptor', 'is_descriptor', 'fields', 'field_descriptors', 'calcsize',
-    'EByteOrder', 'EBitOrder', 'EBaseUnits',
-    'baseunits', 'byteorder', 'bitorder',
-    'field', 'Field', 'is_field',
+    'descriptor', 'is_descriptor', 'fields', 'asdict', 'calcsize',
+    'baseunits', 'byteorder', 'bitorder', 'field', 'Field', 'is_field',
+    'field_descriptors',
     'BinFieldDescriptor', 'get_field_descriptor', 'set_field_descriptor',
     'BASEUNITS_ATTR_NAME', 'BYTEORDER_ATTR_NAME', 'BITORDER_ATTR_NAME',
-    'DECODER_ATTR_NAME', 'METADATA_KEY',
+    'METADATA_KEY',
 ]
 
 
 BASEUNITS_ATTR_NAME = '__bpack_baseunits__'
 BYTEORDER_ATTR_NAME = '__bpack_byteorder__'
 BITORDER_ATTR_NAME = '__bpack_bitorder__'
-DECODER_ATTR_NAME = '__bpack_decoder__'
+SIZE_ATTR_NAME = '__bpack_size__'
 METADATA_KEY = '__bpack_metadata__'
 
 
-class EBaseUnits(enum.Enum):
-    """Base units used to specify size and offset parameters in descriptors."""
-
-    BITS = 'bits'
-    BYTES = 'bytes'
+class DescriptorConsistencyError(ValueError):
+    pass
 
 
-class EByteOrder(enum.Enum):
-    """Enumeration for byte order (endianess).
+class NotFieldDescriptorError(TypeError):
+    pass
 
-    .. note::
 
-        the :data:`EByteOrder.DEFAULT` is equivalent to
-        :data:`EByteOrder.LITTLE` for binary structures having
-        :data:`EBaseUnits.BYTE` base units, and :data:`EByteOrder.BIG`
-        for binary structures having :data:`EBaseUnits.BYTE` base units.
+def _resolve_type(type_):
+    """Remove type annotations.
+
+    Replace :class:`typing.Annotated` types with the corresponding
+    not-annotated ones.
     """
-
-    BIG = '>'
-    LITTLE = '<'
-    NATIVE = '='
-    DEFAULT = ''
+    if bpack.utils.is_sequence_type(type_):
+        etype = bpack.utils.effective_type(type_)
+        rtype = copy.copy(type_)
+        rtype.__args__ = (etype,)
+    elif bpack.typing.is_annotated(type_):
+        rtype = bpack.utils.effective_type(type_)
+    else:
+        rtype = type_
+    return rtype
 
 
-class EBitOrder(enum.Enum):
-    """Enumeration for bit order."""
-
-    MSB = '>'
-    LSB = '<'
-    DEFAULT = ''
-
-
-# TODO: converters (TBD, or in decoder)
 @dataclasses.dataclass
 class BinFieldDescriptor:
     """Descriptor for bpack fields."""
@@ -163,11 +156,44 @@ class BinFieldDescriptor:
         repeat = self.repeat if self.repeat is not None else 1
         return self.size * repeat
 
+    @staticmethod
+    def _is_compatible_param(old, new):
+        if old is not None and new is not None and old != new:
+            return False
+        return True
+
+    def update_from_type(self, type_: Type):
+        if self.type is not None:
+            raise TypeError('the type attribute is already set')
+        if bpack.typing.is_annotated(type_):
+            _, params = bpack.typing.get_args(type_)
+            valid = True
+            if not self._is_compatible_param(self.size, params.size):
+                valid = False
+            if not self._is_compatible_param(self.signed, params.signed):
+                valid = False
+            if not valid:
+                raise DescriptorConsistencyError(
+                    f'type string "{params}" is incompatible with the '
+                    f'field descriptor {self}')
+
+            self.type = params.type
+            if self.signed is None:
+                self.signed = params.signed
+            if self.size is None:
+                self.size = params.size
+        elif bpack.utils.is_sequence_type(type_):
+            etype = bpack.utils.effective_type(type_, keep_annotations=True)
+            self.update_from_type(etype)
+            self.type = _resolve_type(type_)
+        else:
+            self.type = type_
+
 
 Field = dataclasses.Field
 
 
-def field(*, size: int, offset: Optional[int] = None,
+def field(*, size: Optional[int] = None, offset: Optional[int] = None,
           signed: Optional[bool] = None, repeat: Optional[int] = None,
           metadata=None, **kwargs) -> Field:
     """Initialize a field descriptor.
@@ -196,19 +222,19 @@ def _update_field_metadata(field_, **kwargs):
     return field_
 
 
-def get_field_descriptor(field: Field,
+def get_field_descriptor(field: Field,                                  # noqa
                          validate: bool = True) -> BinFieldDescriptor:
     """Return the field descriptor attached to a :class:`Field`."""
     if not is_field(field):
-        raise TypeError(f'not a field descriptor: {field}')
+        raise NotFieldDescriptorError(f'not a field descriptor: {field}')
     field_descr = BinFieldDescriptor(**field.metadata[METADATA_KEY])
-    field_descr.type = field.type
+    field_descr.update_from_type(field.type)
     if validate:
         field_descr.validate()
     return field_descr
 
 
-def set_field_descriptor(field: Field, descriptor: BinFieldDescriptor,
+def set_field_descriptor(field: Field, descriptor: BinFieldDescriptor,  # noqa
                          validate: bool = True) -> Field:
     """Set the field metadata according to the specified descriptor."""
     if validate:
@@ -218,7 +244,11 @@ def set_field_descriptor(field: Field, descriptor: BinFieldDescriptor,
         if v is not None
     }
     type_ = field_descr_metadata.pop('type', None)
-    if type_ != field.type:
+    if bpack.typing.is_annotated(field.type):
+        field_type, _ = bpack.typing.get_args(field.type)
+    else:
+        field_type = field.type
+    if type_ != _resolve_type(field_type):
         raise TypeError(
             f'type mismatch between BinFieldDescriptor.type ({type_!r}) and '
             f'filed.type ({field.type!r})')
@@ -229,10 +259,6 @@ def set_field_descriptor(field: Field, descriptor: BinFieldDescriptor,
     return field
 
 
-class DescriptorConsistencyError(ValueError):
-    pass
-
-
 _DEFAULT_SIZE_MAP = {
     EBaseUnits.BYTES: {
         bool: 1,
@@ -247,7 +273,8 @@ _DEFAULT_SIZE_MAP = {
 }
 
 
-def _get_default_size(type_, baseunits: EBaseUnits) -> Union[int, None]:
+def _get_default_size(type_,
+                      baseunits: EBaseUnits) -> Union[int, None]:       # noqa
     if is_descriptor(type_):
         return calcsize(type_, baseunits)
 
@@ -277,14 +304,30 @@ def _get_default_size(type_, baseunits: EBaseUnits) -> Union[int, None]:
     #     else:
     #         return None
 
-    return _DEFAULT_SIZE_MAP[baseunits].get(etype)
+    return _DEFAULT_SIZE_MAP[baseunits].get(etype)                      # noqa
+
+
+def _get_effective_byteorder(byteorder: EByteOrder,                     # noqa
+                             baseunits: EBaseUnits) -> EByteOrder:      # noqa
+    byteorder = EByteOrder(byteorder)                                   # noqa
+    effective_byteorder = byteorder
+    if baseunits is EBaseUnits.BYTES:
+        if byteorder in {EByteOrder.NATIVE, EByteOrder.DEFAULT}:
+            effective_byteorder = EByteOrder.get_native()
+    else:
+        if byteorder is EByteOrder.DEFAULT:
+            effective_byteorder = EByteOrder.BE
+        elif byteorder in EByteOrder.NATIVE:
+            effective_byteorder = EByteOrder.get_native()
+    return effective_byteorder
 
 
 @classdecorator
 def descriptor(cls, *, size: Optional[int] = None,
-               byteorder: Union[str, EByteOrder] = EByteOrder.DEFAULT,
-               bitorder: Optional[Union[str, EBitOrder]] = None,
-               baseunits: EBaseUnits = EBaseUnits.BYTES):
+               byteorder: Union[str, EByteOrder] = EByteOrder.DEFAULT,  # noqa
+               bitorder: Optional[Union[str, EBitOrder]] = None,        # noqa
+               baseunits: EBaseUnits = EBaseUnits.BYTES,                # noqa
+               **kwargs):
     """Class decorator to define descriptors for binary records.
 
     It converts a dataclass into a descriptor object for binary records.
@@ -292,15 +335,39 @@ def descriptor(cls, *, size: Optional[int] = None,
     * ensures that all fields are :class:`bpack.descriptor.Field` descriptors
     * offsets are automatically computed if necessary
     * consistency checks on offsets and sizes are performed
-    * the ``__len__`` special method is added (returning always the
-      record size in bytes).
+
+    :param cls:
+        class to be decorated
+    :param size:
+        the size (expressed in *base units*) of the binary record
+    :param byteorder:
+        the byte-order of the binary record
+    :param bitorder:
+        the bit-order of the binary record
+        (must be ``None`` if the *base units* are bytes)
+    :param baseunits:
+        the base units (:data:`bpack.enums.EBaseUnits.BITS` or
+        :data:`bpack.enums.EBaseUnits.BYTES`) used to specify the
+        binary record descriptor
+
+    It is also possible to specify as additional keyword arguments all the
+    parameters accepted by :func:`dataclasses.dataclass`.
     """
-    baseunits = EBaseUnits(baseunits)
+    baseunits = EBaseUnits(baseunits)                                   # noqa
+    byteorder = EByteOrder(byteorder)                                   # noqa
+
+    if dataclasses.is_dataclass(cls):
+        warnings.warn('the explicit use of dataclasses is deprecated',
+                      category=DeprecationWarning)
+    else:
+        cls = dataclasses.dataclass(cls, **kwargs)
+
     fields_ = dataclasses.fields(cls)
 
     # Initialize to a dummy value with initial offset + size = 0
     prev_field_descr = BinFieldDescriptor(size=None, offset=0)
     prev_field_descr.size = 0  # trick to bypass checks on BinFieldDescriptor
+    content_size = 0
 
     for idx, field_ in enumerate(fields_):
         assert isinstance(field_, Field)
@@ -308,12 +375,24 @@ def descriptor(cls, *, size: Optional[int] = None,
         # NOTE: this is ensured by dataclasses but not by attr
         assert field_.type is not None
 
+        if bpack.typing.is_annotated(field_.type):
+            # check byteorder
+            _, params = bpack.typing.get_args(field_.type)
+            if params.byteorder:
+                effective_byteorder = _get_effective_byteorder(byteorder,
+                                                               baseunits)
+                if params.byteorder != effective_byteorder:
+                    raise DescriptorConsistencyError(
+                        f'the byteorder of field {field_.name} '
+                        f'("{params.byteorder}" is not consistent with the '
+                        f'descriptor byteorder ({byteorder}) )')
+
         try:
             field_descr = get_field_descriptor(field_, validate=False)
-        except TypeError:
+        except NotFieldDescriptorError:
             field_descr = BinFieldDescriptor()
             if isinstance(field_, Field):
-                field_descr.type = field_.type
+                field_descr.update_from_type(field_.type)
 
         if field_descr.size is None:
             field_descr.size = _get_default_size(field_descr.type, baseunits)
@@ -338,9 +417,8 @@ def descriptor(cls, *, size: Optional[int] = None,
 
         set_field_descriptor(field_, field_descr)
         prev_field_descr = field_descr
+        content_size += field_descr.size
 
-    content_size = sum(
-        get_field_descriptor(field_).size for field_ in fields_)
     field_descr = get_field_descriptor(fields_[-1])
     auto_size = field_descr.offset + field_descr.total_size
     assert auto_size >= content_size  # this should be already checked above
@@ -357,40 +435,32 @@ def descriptor(cls, *, size: Optional[int] = None,
     if baseunits is EBaseUnits.BITS:
         if size % 8 != 0:
             warnings.warn('bit struct not aligned to bytes')
-        size = math.ceil(size / 8)
-
-    setattr(cls, BASEUNITS_ATTR_NAME, baseunits)
-    setattr(cls, BYTEORDER_ATTR_NAME, EByteOrder(byteorder))
 
     if baseunits is not EBaseUnits.BITS and bitorder is not None:
         raise ValueError(
             'it is not possible to specify the "bitorder" '
             'if "baseunits" is not "BITS"')
     elif baseunits is EBaseUnits.BITS and bitorder is None:
-        bitorder = EBitOrder.DEFAULT
+        bitorder = EBitOrder.DEFAULT                                    # noqa
 
+    setattr(cls, BASEUNITS_ATTR_NAME, baseunits)
+    setattr(cls, BYTEORDER_ATTR_NAME, byteorder)
     setattr(cls, BITORDER_ATTR_NAME,
             EBitOrder(bitorder) if bitorder is not None else None)
-
-    get_len_func = bpack.utils.create_fn(
-        name='__len__',
-        args=tuple(),
-        body=['return size'],
-        locals={'size': size},
-    )
-    get_len_func = staticmethod(get_len_func)
-    get_len_func.__doc__ = "Return the record size in bytes"
-    assert not hasattr(cls, '__len__')
-    bpack.utils.set_new_attribute(cls, '__len__', get_len_func)
+    setattr(cls, SIZE_ATTR_NAME, size)
 
     return cls
 
 
+def fields(obj) -> Sequence[Field]:
+    """Return a tuple describing the fields of this descriptor."""
+    return dataclasses.fields(obj)                                      # noqa
+
+
 def is_descriptor(obj) -> bool:
     """Return true if ``obj`` is a descriptor or a descriptor instance."""
     try:
-        return (hasattr(obj, BASEUNITS_ATTR_NAME) and
-                is_field(dataclasses.fields(obj)[0]))
+        return hasattr(obj, BASEUNITS_ATTR_NAME) and is_field(fields(obj)[0])
     except (TypeError, ValueError):
         # dataclass.fields(...) --> TypeError
         # attr.fields(...)      --> NotAnAttrsClassError(ValueError)
@@ -400,16 +470,40 @@ def is_descriptor(obj) -> bool:
         return False
 
 
-def calcsize(obj, units: EBaseUnits = EBaseUnits.BYTES) -> int:
+def asdict(obj, *, dict_factory=dict) -> dict:
+    """Return the fields of a record as a new dictionary.
+
+     The returned dictionary maps field names to field values.
+
+    If given, 'dict_factory' will be used instead of built-in dict.
+    The function applies recursively to field values that are
+    dataclass instances. This will also look into built-in containers:
+    tuples, lists, and dicts.
+    """
+    return dataclasses.asdict(obj, dict_factory=dict_factory)
+
+
+def calcsize(obj, units: Optional[EBaseUnits] = None) -> int:
     """Return the size of the ``obj`` record.
 
-    By default the returned size is expressed in bytes.
+    If the *units* parameter is not specified (default) then the returned
+    *size* is expressed in the same *base units* of the descriptor.
     """
     if not is_descriptor(obj):
         raise TypeError(f'{obj!r} is not a descriptor')
 
-    bytes_to_units = 1 if EBaseUnits(units) is EBaseUnits.BYTES else 8
-    return obj.__len__() * bytes_to_units
+    size = getattr(obj, SIZE_ATTR_NAME)
+    baseunits_ = getattr(obj, BASEUNITS_ATTR_NAME)
+
+    if units and units is not baseunits_:
+        if units is EBaseUnits.BYTES:
+            # baseunits is BITS and units is BYTES
+            size = math.ceil(size / 8)
+        else:
+            # baseunits is BYTES and units is BITS
+            size *= 8
+
+    return size
 
 
 def baseunits(obj) -> EBaseUnits:
@@ -436,10 +530,8 @@ def bitorder(obj) -> Union[EBitOrder, None]:
         raise TypeError(f'"{obj}" is not a descriptor')
 
 
-fields = dataclasses.fields
-
-
-def field_descriptors(descriptor, pad=False) -> Iterable[BinFieldDescriptor]:
+def field_descriptors(descriptor,                                       # noqa
+                      pad: bool = False) -> Iterable[BinFieldDescriptor]:
     """Return the list of field descriptors for the input record descriptor.
 
     Items are instances of the :class:`BinFieldDescriptor` class describing
@@ -451,7 +543,7 @@ def field_descriptors(descriptor, pad=False) -> Iterable[BinFieldDescriptor]:
     """
     if pad:
         offset = 0
-        for field_ in dataclasses.fields(descriptor):
+        for field_ in fields(descriptor):
             field_descr = get_field_descriptor(field_)
             assert field_descr.offset >= offset
             if field_descr.offset > offset:
@@ -467,5 +559,5 @@ def field_descriptors(descriptor, pad=False) -> Iterable[BinFieldDescriptor]:
             # padding
             yield BinFieldDescriptor(size=size - offset, offset=offset)
     else:
-        for field_ in dataclasses.fields(descriptor):
+        for field_ in fields(descriptor):
             yield get_field_descriptor(field_)
