@@ -2,27 +2,72 @@
 
 import sys
 import enum
-import dataclasses
 
 from typing import List, Sequence
 
 import pytest
 
 import bpack
-import bpack.bs
 import bpack.st
+import bpack.codecs
+
+try:
+    import bpack.bs as bpack_bs
+except ImportError:                                         # pragma: no cover
+    bpack_bs = None
+
+try:
+    import bpack.ba as bpack_ba
+except ImportError:                                         # pragma: no cover
+    bpack_ba = None
+
+try:
+    import bpack.np as bpack_np
+except ImportError:                                         # pragma: no cover
+    bpack_np = None
 
 
-@pytest.mark.parametrize('backend', [bpack.bs, bpack.st])
+skipif = pytest.mark.skipif
+
+BITS_BACKENDS = [
+    pytest.param(bpack_bs, id='bs',
+                 marks=skipif(not bpack_bs, reason='not available')),
+    pytest.param(bpack_ba, id='ba',
+                 marks=skipif(not bpack_ba, reason='not available')),
+]
+BYTES_BACKENDS = [
+    pytest.param(bpack.st, id='st'),
+    pytest.param(bpack_np, id='np',
+                 marks=skipif(not bpack_np, reason='not available')),
+]
+ALL_BACKENDS = BITS_BACKENDS + BYTES_BACKENDS
+
+
+@pytest.mark.parametrize('backend', ALL_BACKENDS)
 def test_backend(backend):
     assert hasattr(backend, 'BACKEND_NAME')
     assert hasattr(backend, 'BACKEND_TYPE')
 
 
+@pytest.mark.parametrize('backend', ALL_BACKENDS)
+def test_attrs(backend):
+    @backend.decoder
+    @bpack.descriptor(baseunits=backend.BACKEND_TYPE)
+    class Record:
+        field_1: int = bpack.field(size=4, default=0)
+        field_2: int = bpack.field(size=4, default=1)
+
+    assert hasattr(Record, bpack.descriptors.BASEUNITS_ATTR_NAME)
+    assert hasattr(Record, bpack.descriptors.BYTEORDER_ATTR_NAME)
+    assert hasattr(Record, bpack.descriptors.BITORDER_ATTR_NAME)
+    assert hasattr(Record, bpack.descriptors.SIZE_ATTR_NAME)
+    assert hasattr(Record, bpack.codecs.CODEC_ATTR_NAME)
+
+
 @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
-                  byteorder=bpack.EByteOrder.BIG,
-                  bitorder=bpack.EBitOrder.MSB)
-@dataclasses.dataclass(frozen=True)
+                  byteorder=bpack.EByteOrder.BE,
+                  bitorder=bpack.EBitOrder.MSB,
+                  frozen=True)
 class BitRecordBeMsb:
     # default (unsigned)
     field_01: bool = bpack.field(size=1, default=True)
@@ -74,9 +119,9 @@ BIT_ENCODED_DATA_BE_MSB = b''.join([
 
 
 @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
-                  byteorder=bpack.EByteOrder.LITTLE,
-                  bitorder=bpack.EBitOrder.MSB)
-@dataclasses.dataclass(frozen=True)
+                  byteorder=bpack.EByteOrder.LE,
+                  bitorder=bpack.EBitOrder.MSB,
+                  frozen=True)
 class BitRecordLeMsb:
     # default (unsigned)
     field_01: bool = bpack.field(size=1, default=True)
@@ -128,9 +173,9 @@ BIT_ENCODED_DATA_LE_MSB = b''.join([
 
 
 @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
-                  byteorder=bpack.EByteOrder.BIG,
-                  bitorder=bpack.EBitOrder.LSB)
-@dataclasses.dataclass(frozen=True)
+                  byteorder=bpack.EByteOrder.BE,
+                  bitorder=bpack.EBitOrder.LSB,
+                  frozen=True)
 class BitRecordBeLsb:
     # default (unsigned)
     field_01: bool = bpack.field(size=1, default=True)
@@ -182,9 +227,9 @@ BIT_ENCODED_DATA_BE_LSB = b''.join([
 
 
 @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
-                  byteorder=bpack.EByteOrder.LITTLE,
-                  bitorder=bpack.EBitOrder.LSB)
-@dataclasses.dataclass(frozen=True)
+                  byteorder=bpack.EByteOrder.LE,
+                  bitorder=bpack.EBitOrder.LSB,
+                  frozen=True)
 class BitRecordLeLsb:
     # default (unsigned)
     field_01: bool = bpack.field(size=1, default=True)
@@ -236,8 +281,8 @@ BIT_ENCODED_DATA_LE_LSB = b''.join([
 
 
 @bpack.descriptor(baseunits=bpack.EBaseUnits.BYTES,
-                  byteorder=bpack.EByteOrder.BIG)
-@dataclasses.dataclass(frozen=True)
+                  byteorder=bpack.EByteOrder.BE,
+                  frozen=True)
 class ByteRecordBe:
     field_01: bool = bpack.field(size=1, default=False)
 
@@ -308,8 +353,8 @@ BYTE_ENCODED_DATA_BE = bytes([
 
 
 @bpack.descriptor(baseunits=bpack.EBaseUnits.BYTES,
-                  byteorder=bpack.EByteOrder.LITTLE)
-@dataclasses.dataclass(frozen=True)
+                  byteorder=bpack.EByteOrder.LE,
+                  frozen=True)
 class ByteRecordLe:
     field_01: bool = bpack.field(size=1, default=False)
 
@@ -371,7 +416,7 @@ BYTE_ENCODED_DATA_LE = bytes([
     0b01100001, 0b01100010, 0b01100011,                 # b'abc'    bytes
     0b01000001, 0b01000010, 0b01000011,                 # 'ABC'     str
 
-    # 4 padding bytes ([36:40])
+    # 4 padding bytes ([66:70])
     0b01111000, 0b01111000, 0b01111000, 0b01111000,     # b'xxxx'   bytes
 
     # field_20
@@ -380,15 +425,32 @@ BYTE_ENCODED_DATA_LE = bytes([
 
 
 @pytest.mark.parametrize(
-    'backend, Record, encoded_data, decoded_data',
-    [(bpack.st, ByteRecordBe, BYTE_ENCODED_DATA_BE, ByteRecordBe()),
-     (bpack.st, ByteRecordLe, BYTE_ENCODED_DATA_LE, ByteRecordLe()),
-     (bpack.bs, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, BitRecordBeMsb()),
-     (bpack.bs, BitRecordLeMsb, BIT_ENCODED_DATA_LE_MSB, BitRecordLeMsb()),
-     (bpack.bs, BitRecordBeLsb, BIT_ENCODED_DATA_BE_LSB, BitRecordBeLsb()),
-     (bpack.bs, BitRecordLeLsb, BIT_ENCODED_DATA_LE_LSB, BitRecordLeLsb())],
-    ids=['st BE', 'st LE', 'bs BE MSB', 'bs LE MSB', 'bs BE LSB', 'bs LE LSB'])
-def test_decoder(backend, Record, encoded_data, decoded_data):  # noqa
+    'backend, Record, encoded_data',
+    [pytest.param(bpack.st, ByteRecordBe, BYTE_ENCODED_DATA_BE, id='st BE'),
+     pytest.param(bpack.st, ByteRecordLe, BYTE_ENCODED_DATA_LE, id='st LE'),
+     pytest.param(
+         bpack_np, ByteRecordBe, BYTE_ENCODED_DATA_BE, id='np BE',
+         marks=pytest.mark.skipif(not bpack_np, reason='not available')),
+     pytest.param(
+         bpack_np, ByteRecordLe, BYTE_ENCODED_DATA_LE, id='np LE',
+         marks=pytest.mark.skipif(not bpack_np, reason='not available')),
+     pytest.param(
+         bpack_bs, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, id='bs BE MSB',
+         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
+     pytest.param(
+         bpack_bs, BitRecordLeMsb, BIT_ENCODED_DATA_LE_MSB, id='bs LE MSB',
+         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
+     pytest.param(
+         bpack_bs, BitRecordBeLsb, BIT_ENCODED_DATA_BE_LSB, id='bs BE LSB',
+         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
+     pytest.param(
+         bpack_bs, BitRecordLeLsb, BIT_ENCODED_DATA_LE_LSB, id='bs LE LSB',
+         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
+     pytest.param(
+         bpack_ba, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, id='ba BE MSB',
+         marks=pytest.mark.skipif(not bpack_ba, reason='not available'))])
+def test_decoder(backend, Record, encoded_data):                        # noqa
+    decoded_data = Record()                                             # noqa
     decoder = backend.Decoder(Record)
     assert hasattr(decoder, 'baseunits')
     assert decoder.baseunits is bpack.baseunits(Record)
@@ -397,25 +459,41 @@ def test_decoder(backend, Record, encoded_data, decoded_data):  # noqa
 
 
 @pytest.mark.parametrize(
-    'backend, Record, encoded_data, decoded_data',
-    [(bpack.st, ByteRecordBe, BYTE_ENCODED_DATA_BE, ByteRecordBe()),
-     (bpack.st, ByteRecordLe, BYTE_ENCODED_DATA_LE, ByteRecordLe()),
-     (bpack.bs, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, BitRecordBeMsb()),
-     (bpack.bs, BitRecordLeMsb, BIT_ENCODED_DATA_LE_MSB, BitRecordLeMsb()),
-     (bpack.bs, BitRecordBeLsb, BIT_ENCODED_DATA_BE_LSB, BitRecordBeLsb()),
-     (bpack.bs, BitRecordLeLsb, BIT_ENCODED_DATA_LE_LSB, BitRecordLeLsb())],
-    ids=['st BE', 'st LE', 'bs BE MSB', 'bs LE MSB', 'bs BE LSB', 'bs LE LSB'])
-def test_decoder_func(backend, Record, encoded_data, decoded_data):  # noqa
+    'backend, Record, encoded_data',
+    [pytest.param(bpack.st, ByteRecordBe, BYTE_ENCODED_DATA_BE, id='st BE'),
+     pytest.param(bpack.st, ByteRecordLe, BYTE_ENCODED_DATA_LE, id='st LE'),
+     pytest.param(
+         bpack_np, ByteRecordBe, BYTE_ENCODED_DATA_BE, id='np BE',
+         marks=pytest.mark.skipif(not bpack_np, reason='not available')),
+     pytest.param(
+         bpack_np, ByteRecordLe, BYTE_ENCODED_DATA_LE, id='np LE',
+         marks=pytest.mark.skipif(not bpack_np, reason='not available')),
+     pytest.param(
+         bpack_bs, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, id='bs BE MSB',
+         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
+     pytest.param(
+         bpack_bs, BitRecordLeMsb, BIT_ENCODED_DATA_LE_MSB, id='bs LE MSB',
+         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
+     pytest.param(
+         bpack_bs, BitRecordBeLsb, BIT_ENCODED_DATA_BE_LSB, id='bs BE LSB',
+         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
+     pytest.param(
+         bpack_bs, BitRecordLeLsb, BIT_ENCODED_DATA_LE_LSB, id='bs LE LSB',
+         marks=pytest.mark.skipif(not bpack_bs, reason='not available')),
+     pytest.param(
+         bpack_ba, BitRecordBeMsb, BIT_ENCODED_DATA_BE_MSB, id='ba BE MSB',
+         marks=pytest.mark.skipif(not bpack_ba, reason='not available'))])
+def test_decoder_func(backend, Record, encoded_data):                   # noqa
+    decoded_data = Record()                                             # noqa
     record_type = backend.decoder(Record)
-    record = record_type.from_bytes(encoded_data)
+    record = record_type.frombytes(encoded_data)
     assert record == decoded_data
 
 
-@pytest.mark.parametrize('backend', [bpack.bs], ids=['bs'])
+@pytest.mark.parametrize('backend', BITS_BACKENDS)
 def test_bit_decoder_decorator(backend):
     @backend.decoder
-    @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS)
-    @dataclasses.dataclass(frozen=True)
+    @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS, frozen=True)
     class Record:
         field_01: bool = bpack.field(size=1, default=True)
         field_02: int = bpack.field(size=3, default=4)
@@ -428,7 +506,7 @@ def test_bit_decoder_decorator(backend):
 
     decoded_data = Record()
     encoded_data = BIT_ENCODED_DATA_BE_MSB[:bpack.calcsize(Record)]
-    record = Record.from_bytes(encoded_data)
+    record = Record.frombytes(encoded_data)
 
     assert record.field_01 == decoded_data.field_01
     assert record.field_02 == decoded_data.field_02
@@ -439,106 +517,105 @@ def test_bit_decoder_decorator(backend):
     assert record.field_08 == decoded_data.field_08
 
 
-@pytest.mark.parametrize('backend', [bpack.st], ids=['st'])
+@pytest.mark.parametrize('backend', BYTES_BACKENDS)
 def test_byte_decoder_decorator(backend):
     @backend.decoder
     @bpack.descriptor(baseunits=bpack.EBaseUnits.BYTES,
-                      byteorder=bpack.EByteOrder.BIG)
-    @dataclasses.dataclass(frozen=True)
+                      byteorder=bpack.EByteOrder.BE,
+                      frozen=True)
     class Record:
         field_1: int = bpack.field(size=1, default=1)
         field_2: int = bpack.field(size=2, default=2)
 
     decoded_data = Record()
     encoded_data = bytes([0b00000001, 0b00000000, 0b00000010])
-    record = Record.from_bytes(encoded_data)
+    record = Record.frombytes(encoded_data)
 
     assert record.field_1 == decoded_data.field_1
     assert record.field_2 == decoded_data.field_2
 
 
-@pytest.mark.parametrize('backend', [bpack.bs, bpack.st], ids=['bs', 'st'])
+@pytest.mark.parametrize('backend', ALL_BACKENDS)
 def test_unsupported_type(backend):
     class CustomType:
         pass
 
     with pytest.raises(TypeError):
         @backend.decoder
-        @bpack.descriptor(baseunits=backend.Decoder.baseunits)
-        @dataclasses.dataclass(frozen=True)
-        class Record:
+        @bpack.descriptor(baseunits=backend.Decoder.baseunits, frozen=True)
+        class Record:                                                   # noqa
             field_1: CustomType = bpack.field(size=8)
 
 
-@pytest.mark.parametrize('backend', [bpack.st], ids=['st'])
+@pytest.mark.parametrize('backend', BYTES_BACKENDS)
 def test_byte_decoder_native_byteorder(backend):
     size = 4
     value = 1
 
     @backend.decoder
-    @bpack.descriptor(byteorder=bpack.EByteOrder.NATIVE)
-    @dataclasses.dataclass(frozen=True)
+    @bpack.descriptor(byteorder=bpack.EByteOrder.NATIVE, frozen=True)
     class Record:
         field_1: int = bpack.field(size=size, default=value)
 
     data = value.to_bytes(size, sys.byteorder)
-    assert Record.from_bytes(data) == Record()
+    assert Record.frombytes(data) == Record()
 
 
-@pytest.mark.parametrize('backend', [bpack.bs], ids=['bs'])
+@pytest.mark.parametrize(
+    'backend',
+    [pytest.param(bpack_bs, id='bs',
+                  marks=skipif(not bpack_bs, reason='not available'))])
 def test_bit_decoder_native_byteorder(backend):
     size = 8
     value = 1
 
     @backend.decoder
     @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
-                      byteorder=bpack.EByteOrder.NATIVE)
-    @dataclasses.dataclass(frozen=True)
+                      byteorder=bpack.EByteOrder.NATIVE,
+                      frozen=True)
     class Record:
         field_1: int = bpack.field(size=8, default=1)
 
     data = value.to_bytes(size, sys.byteorder)
-    assert Record.from_bytes(data) == Record()
+    assert Record.frombytes(data) == Record()
 
 
-@pytest.mark.parametrize('backend', [bpack.bs], ids=['bs'])
+@pytest.mark.parametrize('backend', BITS_BACKENDS)
 def test_bit_decoder_default_byteorder(backend):
     size = 8
     value = 1
 
     @backend.decoder
     @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS,
-                      byteorder=bpack.EByteOrder.DEFAULT)
-    @dataclasses.dataclass(frozen=True)
+                      byteorder=bpack.EByteOrder.DEFAULT,
+                      frozen=True)
     class Record:
         field_1: int = bpack.field(size=8, default=1)
 
     # default byte order is big for bit descriptors
     data = value.to_bytes(size // 8, 'big')
-    assert Record.from_bytes(data) == Record()
+    assert Record.frombytes(data) == Record()
 
 
-@pytest.mark.parametrize('backend', [bpack.bs], ids=['bs'])
+@pytest.mark.parametrize('backend', BITS_BACKENDS)
 def test_wrong_baseunits_bit(backend):
     with pytest.raises(ValueError):
         @backend.decoder
         @bpack.descriptor(baseunits=bpack.EBaseUnits.BYTES)
-        @dataclasses.dataclass
-        class Record:
+        class Record:                                                   # noqa
             field_1: int = bpack.field(size=8, default=1)
 
 
-@pytest.mark.parametrize('backend', [bpack.st], ids=['st'])
+@pytest.mark.parametrize('backend', BYTES_BACKENDS)
 def test_wrong_baseunits_byte(backend):
     with pytest.raises(ValueError):
         @backend.decoder
         @bpack.descriptor(baseunits=bpack.EBaseUnits.BITS)
-        @dataclasses.dataclass
-        class Record:
+        class Record:                                                   # noqa
             field_1: int = bpack.field(size=8, default=1)
 
 
-@pytest.mark.parametrize('backend', [bpack.st, bpack.bs], ids=['st', 'bs'])
+@pytest.mark.parametrize('backend', ALL_BACKENDS)
 def test_enum_decoding_bytes(backend):
     class EStrEnumType(enum.Enum):
         A = 'a'
@@ -578,7 +655,6 @@ def test_enum_decoding_bytes(backend):
 
     @backend.decoder
     @bpack.descriptor(baseunits=backend.Decoder.baseunits, bitorder=bitorder)
-    @dataclasses.dataclass
     class Record:
         field_1: EStrEnumType = bpack.field(size=ssize, default=EStrEnumType.A)
         field_2: EBytesEnumType = bpack.field(size=ssize,
@@ -587,11 +663,15 @@ def test_enum_decoding_bytes(backend):
         field_4: EFlagEnumType = bpack.field(size=isize,
                                              default=EFlagEnumType.A)
 
-    record = Record.from_bytes(encoded_data)
+    record = Record.frombytes(encoded_data)
     assert record == Record()
 
 
-@pytest.mark.parametrize('backend', [bpack.st, bpack.bs], ids=['st', 'bs'])
+@pytest.mark.parametrize(
+    'backend',
+    [pytest.param(bpack.st, id='st'),
+     pytest.param(bpack_bs, id='bs',
+                  marks=skipif(not bpack_bs, reason='not available'))])
 def test_sequence(backend):
     if backend.Decoder.baseunits is bpack.EBaseUnits.BYTES:
         bitorder = None
@@ -606,13 +686,12 @@ def test_sequence(backend):
 
     @backend.decoder
     @bpack.descriptor(baseunits=backend.Decoder.baseunits, bitorder=bitorder)
-    @dataclasses.dataclass
     class Record:
         field_1: List[int] = bpack.field(size=size, repeat=repeat)
         field_2: Sequence[int] = bpack.field(size=size, repeat=repeat)
 
     ref_record = Record([3, 3], (4, 4))
-    record = Record.from_bytes(encoded_data)
+    record = Record.frombytes(encoded_data)
     assert record == ref_record
 
     assert type(record.field_1) is list
@@ -623,7 +702,13 @@ def test_sequence(backend):
         assert field.type == sequence_type
 
 
-@pytest.mark.parametrize('backend', [bpack.bs, bpack.st], ids=['bs', 'st'])
+@pytest.mark.parametrize(
+    'backend',
+    [pytest.param(bpack.st, id='st'),
+     pytest.param(bpack_bs, id='bs',
+                  marks=skipif(not bpack_bs, reason='not available')),
+     pytest.param(bpack_np, id='np',
+                  marks=skipif(not bpack_np, reason='not available'))])
 class TestNestedRecord:
     @staticmethod
     def get_encoded_data(baseunits):
@@ -645,36 +730,32 @@ class TestNestedRecord:
 
         @backend.decoder
         @bpack.descriptor(baseunits=backend.Decoder.baseunits)
-        @dataclasses.dataclass
         class Record:
             field_1: int = bpack.field(size=4, default=1)
             field_2: int = bpack.field(size=4, default=2)
 
         @backend.decoder
         @bpack.descriptor(baseunits=backend.Decoder.baseunits)
-        @dataclasses.dataclass
         class NestedRecord:
             field_1: int = bpack.field(size=4, default=0)
             field_2: Record = Record()
             field_3: int = bpack.field(size=4, default=3)
 
-        assert NestedRecord.from_bytes(encoded_data) == NestedRecord()
+        assert NestedRecord.frombytes(encoded_data) == NestedRecord()
 
     def test_nested_record(self, backend):
         encoded_data = self.get_encoded_data(backend.Decoder.baseunits)
 
         @bpack.descriptor(baseunits=backend.Decoder.baseunits)
-        @dataclasses.dataclass
         class Record:
             field_1: int = bpack.field(size=4, default=1)
             field_2: int = bpack.field(size=4, default=2)
 
         @backend.decoder
         @bpack.descriptor(baseunits=backend.Decoder.baseunits)
-        @dataclasses.dataclass
         class NestedRecord:
             field_1: int = bpack.field(size=4, default=0)
             field_2: Record = Record()
             field_3: int = bpack.field(size=4, default=3)
 
-        assert NestedRecord.from_bytes(encoded_data) == NestedRecord()
+        assert NestedRecord.frombytes(encoded_data) == NestedRecord()
